import numpy as np

from hardware_communicator.message_interpreter.basic_interpreter import (
    AbstractInterpreter,
    SendItem,
)


def bytearray_to_dtype(byte_array, dtype):
    new = np.array(byte_array)
    if dtype == str:
        new = b''.join(new).decode('latin-1')
    else:
        target_size = np.dtype(dtype).itemsize
        size = len(new.tostring())
        new = np.zeros(size % target_size, dtype=np.uint8).tostring() + new.tostring()
        new = np.frombuffer(new, dtype=dtype)
    return new


class BinaryInterpreter(AbstractInterpreter):
    def __init__(self):
        self.queries = {}

    def encode_data(self, data):
        return SendItem(data=bytearray(data))

    def decode_data(self, data, target):
        pass


class StartKeyDataEndInterpreter(BinaryInterpreter):
    def __init__(
            self,
            send_start_code,
            send_end_code,
            receive_start_code=None,
            receive_end_code=None,
    ):
        super().__init__()
        if receive_start_code is None:
            receive_start_code = send_start_code
        if receive_end_code is None:
            receive_end_code = send_end_code

        if isinstance(send_start_code, str):
            send_start_code = send_start_code.encode()
        if isinstance(send_end_code, str):
            send_end_code = send_end_code.encode()
        if isinstance(receive_start_code, str):
            receive_start_code = receive_start_code.encode()
        if isinstance(receive_end_code, str):
            receive_end_code = receive_end_code.encode()

        if isinstance(send_start_code, int):
            send_start_code = [send_start_code]
        if isinstance(send_end_code, int):
            send_end_code = [send_end_code]
        if isinstance(receive_start_code, int):
            receive_start_code = [receive_start_code]
        if isinstance(receive_end_code, int):
            receive_end_code = [receive_end_code]

        self.send_end_code = [x for x in send_end_code]
        self.send_start_code = [x for x in send_start_code]
        self.receive_start_code = [x for x in receive_start_code]
        self.receive_end_code = [x for x in receive_end_code]

    def encode_data(self, key, data=None):
        if data is None:
            data = []
        if isinstance(key, str):
            key = [x for x in key.encode()]
        if isinstance(data, str):
            data = [x for x in data.encode()]
        key = list(bytearray(key))

        d = super().encode_data(self.send_start_code + key + data + self.send_end_code)
        return d

    def add_query(
            self,
            name,
            key,
            send_size=0,
            receive_size=0,
            receive_function=None,
            max_receive_size=None,
            receive_dtype=None,
    ):
        if isinstance(key, str):
            key = key.encode()
        if isinstance(key, int):
            key = [key]

        if max_receive_size is None:
            max_receive_size = receive_size
        assert max_receive_size >= receive_size

        if receive_function is None:
            receive_function = print
        self.queries[name] = dict(
            key=key,
            send_size=send_size,
            receive_size=(receive_size, max_receive_size),
            receive_function=receive_function,
            receive_dtype=receive_dtype,
        )

    def prepare_query(self, name, data=None):
        if data is None:
            data = []
        assert name in self.queries, str(name) + " not in queries ("+str(list(self.queries.keys()))+")"
        assert self.queries[name]["send_size"] == len(data)
        return self.encode_data(key=self.queries[name]["key"], data=data)

    def decode_data(self, data, target):
        num_data = [ord(x) for x in data]
        data_start_position = 0
        break_position = 0
        if len(self.receive_start_code) > 0:
            try:
                start_position = num_data.index(self.receive_start_code[0])
                for i, c in enumerate(self.receive_start_code):
                    if c != num_data[i + start_position]:
                        data_start_position = -1
                        break_position = start_position
                        break
                if data_start_position > -1:
                    data_start_position = start_position + len(self.receive_start_code)
            except ValueError:
                return []
        if data_start_position > -1:
            data_end_position = -1
            break_position = data_start_position
            try:
                end_position = num_data.index(self.receive_end_code[0])
                for i, c in enumerate(self.receive_end_code):
                    if c != num_data[i + end_position]:
                        end_position = -1
                        break
                if end_position > -1:
                    data_end_position = end_position
            except:
                pass
            if data_end_position > -1:
                key_and_data = data[data_start_position:data_end_position]
                try:
                    for name, query in self.queries.items():
                        key = query["key"]
                        if key == b"".join(key_and_data[: len(key)]):
                            raw_data = key_and_data[len(key):]
                            receive_size = query["receive_size"]
                            if receive_size[1] >= len(raw_data) >= receive_size[0]:
                                if query["receive_dtype"]:
                                    raw_data = bytearray_to_dtype(
                                        raw_data, query["receive_dtype"]
                                    )
                                try:
                                    query["receive_function"](target, raw_data)
                                except Exception as e:
                                    return data[data_end_position + 1:]
                    return data[data_end_position + 1:]
                except Exception as e:
                    raise e
        return data[break_position:]



import glob
import logging
import sys
import threading
import time

import serial

from hardware_communicator.abstract_communicator import AbstractCommunicator


def get_avalable_serial_ports(ignore=None):
    if ignore is None:
        ignore = []
    if sys.platform.startswith("win"):
        ports = ["COM%s" % (i + 1) for i in range(256)]
    elif sys.platform.startswith("linux") or sys.platform.startswith("cygwin"):
        # this excludes your current terminal "/dev/tty"
        ports = glob.glob("/dev/tty[A-Za-z]*")
    elif sys.platform.startswith("darwin"):
        ports = glob.glob("/dev/tty.*")
    else:
        raise EnvironmentError("Unsupported platform")

    ports = set(ports)

    result = []
    for port in ports.difference(ignore):
        try:
            s = serial.Serial(port)
            s.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass
    return set(result)


PORT_READ_TIME = 0.01


class SerialCommunicator(AbstractCommunicator):
    POSSIBLE_BAUD_RATES = [
        460800,
        230400,
        115200,
        57600,
        38400,
        19200,
        14400,
        9600,
        4800,
        2400,
        1200,
    ]

    def __init__(self,interpreter, port=None, auto_port=True,**kwargs):
        super().__init__(interpreter=interpreter,**kwargs)
        self.connection_checks = []
        self.port = port
        self.connected = False
        self.serial_connection = None
        self.is_open = False
        self.read_buffer = []
        self._work_port_task=None
        self._in_finding = False
        if not hasattr(self, "logger"):
            self.logger = logging.getLogger(self.__class__.__name__)
        if not hasattr(self, "on_connect"):
            self.on_connect = None
        if self.port:
            self.connect_to_port(port)

        if not self.connected:
            if auto_port and self.interpreter:
                def ap():
                    time.sleep(0.5)
                    self.find_port()
                threading.Thread(target=ap,daemon=True).start()


    def find_port(self, excluded_ports=None, retries=3):
        if self.connected:
            return self.port
        if self._in_finding:
            return None
        self._in_finding=True

        if excluded_ports is None:
            excluded_ports = []

        self.logger.info("Check ports"+str(get_avalable_serial_ports(ignore=excluded_ports)))
        for port in get_avalable_serial_ports(ignore=excluded_ports):
            self.connect_to_port(port, retries=retries)
            if self.connected:
                self._in_finding=False
                return self.port
        self._in_finding = False

    def add_connection_check(self, function):
        self.connection_checks.append(function)

    def connect_to_port(self, port, retries=3):
        for i in range(retries):
            self.logger.debug(f'try connecting to port "{port} try {i + 1}/{retries}"')
            for baud in self.POSSIBLE_BAUD_RATES:
                self.logger.debug(f'try connecting to port "{port} with baud {baud}"')
                try:
                    self._open_port(port=port, baud=baud)
                    time.sleep(0.1)
                    check = True
                    for func in self.connection_checks:
                        r = func()
                        if not r:
                            check = False
                            break
                    if check:
                        time.sleep(0.5)
                        self.connected = True
                        break
                    else:
                        self.stop_read(permanently=True)
                except serial.serialutil.SerialException:
                    time.sleep(0.5)
            if self.connected:
                self.port = port
                self.logger.info(f'successfully connected to "{port} with baud {baud}"')
                if self.on_connect:
                    self.on_connect()
                return port

    def _close_port(self):
        self.port = None
        port = None
        if self.serial_connection:
            port = self.serial_connection.port
            self.serial_connection.close()
        if self.is_open:
            self.logger.info("port closed " + port)
        self.serial_connection = None
        self.is_open = False

    def _open_port(self, port, baud):
        self.port = port
        if self.serial_connection or self.is_open:
            self._close_port()
        self.serial_connection = serial.Serial(port, baudrate=baud, timeout=0)
        self.is_open = True
        self._work_port_task = threading.Thread(target=self.work_port,daemon=True)
        self._work_port_task.start()

    def work_port(self):
        while self.is_open:
            try:
                if self.is_open:
                    self._write_to_port()
                    try:
                        c = self.serial_connection.read()
                    except AttributeError as e:
                        c = ""
                    while len(c) > 0:
                        #print(ord(c),c)
                        self.read_buffer.append(c)
                        self.validate_buffer()
                        if not self.is_open:
                            break
                        try:
                            c = self.serial_connection.read()
                        except AttributeError as e:
                            c = ""
                time.sleep(PORT_READ_TIME)
            except Exception as e:
                self.logger.exception(e)
                self.stop_read()
        self.logger.error("work_port stopped")
        self.stop_read()

    def stop_read(self, permanently=False):
        port = None
        baud = None
        if self.serial_connection:
            port = self.serial_connection.port
            baud = self.serial_connection.baudrate
        self._close_port()
        if not permanently and port:
            self._open_port(port=port, baud=baud)

    def detatch(self):
        self.stop_read(permanently=True)

    def write_to_port(self, send_item):
        return self.send_queue.append(send_item)

    def _write_to_port(self):
        #   if(len(self.send_queue)>0):
        # print(self.port, self.send_queue)
        for item in self.send_queue:
            try:
                self.serial_connection.write(list(item.data))
            except Exception as e:
                self.logger.error(f"cannot write {item}")
                raise e
            item.sended(self)

    def validate_buffer(self):
        self.read_buffer = self.interpreter.decode_data(self.read_buffer, self)



import logging

from json_dict import JsonDict

from hardware_communicator.usb_commmunicator import SerialCommunicator


class SerialDevice(SerialCommunicator):
    AVAILABLE_QUERIES = {}

    def __init__(self, port=None, auto_port=True,interpreter=None,**kwargs):
        SerialCommunicator.__init__(self, port=port, auto_port=auto_port,interpreter=interpreter,**kwargs)
        if not hasattr(self, "config"):
            self.config = JsonDict()

    def set_device_property(self, name, value):
        self.config.put("device", "properties", name, value=value)

    def get_device_property(self, name):
        return self.config.get("device", "properties", name, autosave=False)

    def set_device_status(self, name, value):
        self.config.put("device", "status", name, value=value)

    def get_device_status(self, name):
        return self.config.get("device", "status", name, autosave=False)


import functools
import random
import threading
import time


class BackgroundTask:
    def __init__(self, id, method, delay, name=None, max_callings=None,initial_delay=0):
        self._run_thread:threading.Thread = None
        if name is None:
            rff = method
            while isinstance(rff, functools.partial):
                rff = rff.func
            name=rff.__name__
        self.name = name
        self.remain_callings = max_callings
        self.delay = delay
        self.method = method
        self.id = id
        self._last_run = time.time()+initial_delay
        self.running=True

    def check_for_run(self, t=None):
        if self.remain_callings == 0:
            self.stop()
            return None

        if self.delay is None:
            return self.run(t)

        if t is None:
            t = time.time()
        if t >= self._last_run + self.delay:
            return self.run(t)

    def run(self, t=None):
        if t is None:
            t = time.time()
        self._last_run = t
        if self.remain_callings:
            self.remain_callings -= 1
        return self.method()

    def stop(self):
        self.running=False
        try:
            self._run_thread.join(timeout=self.delay+0.2)
        except:
            pass

    def __repr__(self):
        return f"{self.name}({self.delay},{self.remain_callings},{self.id})"

    def start(self,background_sleep_time=None):
        if background_sleep_time is None:
            background_sleep_time = 0.01
        if self._run_thread:
            self.stop()
        def runner():
            while self.running:
                self.check_for_run()
                time.sleep(background_sleep_time)
        self._run_thread = threading.Thread(target=runner,daemon=True)
        self._run_thread.start()

class ThreadedBackgroundTaskManager():
    def __init__(self,background_sleep_time=1):
        self.background_sleep_time = background_sleep_time
        self._background_tasks = {}

    def stop(self):
        for id in self._background_tasks.copy():
            self.stop_task(id)
        self._running = False

    def has_running_tasks(self):
        return any([t.running for t in self._background_tasks.values()])

    def stop_task(self, id):
        if id in self._background_tasks:
            try:
                self._background_tasks[id].stop()
            except:
                pass
            try:
                del self._background_tasks[id]
            except:
                pass

    def register_background_task(self,method, minimum_call_delay=None, max_callings=None,name=None,delay=0):
        task_id = random.randint(1, 10 ** 6)
        while task_id in self._background_tasks:
            task_id = random.randint(1, 10 ** 6)
        if name is None:
            rff = method
            while isinstance(rff, functools.partial):
                rff = rff.func
            name=rff.__name__
        self._background_tasks[task_id] = BackgroundTask(
            id=task_id, method=method, delay=minimum_call_delay, max_callings=max_callings,name=name,initial_delay=delay
        )
        self._background_tasks[task_id].start(self.background_sleep_time)
        return task_id

class AbstractCommunicator(ThreadedBackgroundTaskManager):
    def __init__(self, interpreter,**kwargs):
        super().__init__(**kwargs)
        self._interpreter = None
        self.interpreter=interpreter
        self.send_queue = []

    @property
    def interpreter(self):
        return self._interpreter

    @interpreter.setter
    def interpreter(self, interpreter):
        self._interpreter = interpreter



import os
import sys

import Cython
import setuptools
from Cython.Build import cythonize

with open("requirements.txt") as f:
    required = f.read().splitlines()


def find_pyx(path="."):
    pyx_files = []
    for root, dirs, filenames in os.walk(path):
        for fname in filenames:
            if fname.endswith(".pyx"):
                pyx_files.append(os.path.join(root, fname))
    return pyx_files


extensions = find_pyx()

setup = dict(
    name="hardware_communicator",
    version="0.1",
    author="Julian Kimmig",
    author_email="julian-kimmig@gmx.net",
    description="HardwareCommunicator",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/JulianKimmig/HardwareCommunicator",
    packages=setuptools.find_packages(),
    install_requires=required,
    ext_modules=cythonize(extensions, language_level=3),
    cmdclass={"build_ext": Cython.Build.build_ext},
    package_dir={"hardware_communicator": "hardware_communicator"},
)


def pip_install():
    import subprocess

    subprocess.check_call([sys.executable, "-m", "pip", "install", "."])


if __name__ == "__main__":
    if len(sys.argv) == 1:
        sys.argv.append("build_ext")
        sys.argv.append("--inplace")
        pip_install()
    else:
        setuptools.setup(**setup)



def delete_on_send(item, communicator_target):
    del item


class SendItem:
    def __init__(self, data, on_send=None):
        self.data = data
        if on_send is None:
            on_send = delete_on_send
        self._on_send = on_send

    def sended(self, communicator_target):
        if self in communicator_target.send_queue:
            communicator_target.send_queue.remove(self)
            self._on_send(self, communicator_target)

    def __str__(self):
        return f"{self.__class__.__name__}({self.data})"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data})"


class AbstractInterpreter:
    pass
